# =====================================
# generator=datazen
# version=3.1.4
# hash=bf2c126f5c469e9b0483e934ee695f9e
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "An interface generator for distributed computing."
PKG_NAME = "ifgen"
VERSION = "3.3.0"



"""
A module implementing a data model for ARM CMSIS-SVD 'interrupt' data.
"""

# built-in
from dataclasses import dataclass
from typing import Iterable

# internal
from ifgen.svd.string import StringKeyVal, StringKeyValueMixin


@dataclass
class Interrupt(StringKeyValueMixin):
    """A container for interrupt information."""

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        return [
            StringKeyVal("name", True),
            StringKeyVal("description", False),
            StringKeyVal("value", True),
        ]



"""
A module for common enum-generation utilities.
"""

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.enum.map import enum_from_string_map, enum_to_string_map
from ifgen.generation.interface import GenerateTask


def enum_to_string(
    task: GenerateTask, writer: IndentedFileWriter, use_map: bool
) -> None:
    """
    Implement an enumeration to string method without using the static map.
    """

    writer.write(f'const char *result = "UNKNOWN {task.name}";')

    with writer.padding():
        if not use_map:
            writer.write("switch (instance)")

            with writer.scope(indent=0):
                for enum in task.instance.get("enum", {}):
                    writer.write(f"case {task.name}::{enum}:")
                    with writer.indented():
                        writer.write(f'result = "{enum}";')
                        writer.write("break;")
        else:
            map_inst = f"{task.name}_to_string"
            writer.write(f"if ({map_inst}.contains(instance))")
            with writer.scope():
                writer.write(f"result = {map_inst}.at(instance);")

    writer.write("return result;")


def enum_to_string_function(
    task: GenerateTask,
    writer: IndentedFileWriter,
    use_map: bool,
    definition: bool = False,
) -> None:
    """Generate a method for converting enum instances to strings."""

    if definition:
        with writer.javadoc():
            writer.write(f"Converts {task.name} to a C string.")
            writer.empty()
            writer.write(
                task.command("param[in]", data="instance Value to convert.")
            )
            writer.write(
                task.command(
                    "return",
                    "            A C string representation of the value.",
                )
            )

    if not definition and use_map:
        enum_to_string_map(task, writer)
        writer.empty()

    line = ""
    if not use_map:
        line = "inline "
    line += f"const char *to_string({task.name} instance)"

    if definition and use_map:
        line += ";"

    writer.write(line)

    if definition and use_map:
        return

    with writer.scope():
        enum_to_string(task, writer, use_map)


def enum_from_string(
    task: GenerateTask, writer: IndentedFileWriter, use_map: bool
) -> None:
    """
    Implement an string to enumeration method without using the static map.
    """

    if not use_map:
        writer.write("bool result = false;")

        with writer.padding():
            first = True
            for enum in task.instance.get("enum", {}):
                line = f'if ((result = !strncmp(data, "{enum}", {len(enum)})))'

                if not first:
                    line = "else " + line
                else:
                    first = False

                writer.write(line)
                with writer.scope():
                    writer.write(f"output = {task.name}::{enum};")

    else:
        map_inst = f"{task.name}_from_string"
        writer.write(f"bool result = {map_inst}.contains(data);")

        with writer.padding():
            writer.write("if (result)")
            with writer.scope():
                writer.write(f"output = {map_inst}.at(data);")

    writer.write("return result;")


def string_to_enum_function(
    task: GenerateTask,
    writer: IndentedFileWriter,
    use_map: bool,
    definition: bool = False,
) -> None:
    """Generate a method for converting string instances to enums."""

    if definition:
        with writer.javadoc():
            writer.write(f"Converts a C string to {task.name}.")
            writer.empty()
            writer.write(
                task.command("param[in]", " data   A C string to convert.")
            )
            writer.write(
                task.command(
                    "param[out]",
                    "output The enumeration element to write.",
                )
            )
            writer.write(
                task.command(
                    "return",
                    "           Whether or not the output was written.",
                )
            )

    if not definition and use_map:
        enum_from_string_map(task, writer)
        writer.empty()

    line = ""
    if not use_map:
        line = "inline "

    line += f"bool from_string(const char *data, {task.name} &output)"

    if definition and use_map:
        line += ";"

    writer.write(line)

    if definition and use_map:
        return

    with writer.scope():
        enum_from_string(task, writer, use_map)



"""
Utilities shared between struct methods.
"""

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask


def native_decode(writer: IndentedFileWriter) -> None:
    """Write a buffer decoding method for native byte order."""

    writer.write("auto buf = raw();")
    writer.write("*buf = *buffer;")


def native_encode(writer: IndentedFileWriter) -> None:
    """Write a buffer encoding method for native byte order."""

    writer.write("*buffer = *raw_ro();")


def wrapper_method(
    task: GenerateTask,
    writer: IndentedFileWriter,
    header: bool,
    is_encode: bool = True,
) -> None:
    """Create a generic encode/decode method."""

    method = task.cpp_namespace(
        "encode" if is_encode else "decode", header=header
    )

    if not header:
        writer.c_comment(f"'{method}' defined in header.")
        return

    line_start = f"inline std::size_t {method}("
    line = line_start + ("const " if not is_encode else "") + "Buffer *buffer,"

    writer.write(line)
    line = ""

    line += (
        " " * len(line_start) if header else ""
    ) + "std::endian endianness"
    line += " = std::endian::native"
    line += ")"

    if is_encode:
        line += " const"

    writer.write(line)

    with writer.scope():
        writer.write("std::size_t result = size;")

        with writer.padding():
            writer.write("if (endianness == std::endian::native)")

            with writer.scope():
                # Use trivial implementation for native encode and decode.
                if is_encode:
                    native_encode(writer)
                else:
                    native_decode(writer)

            writer.write("else")

            with writer.scope():
                writer.write(f"result = {method}_swapped(buffer);")

        writer.write("return result;")



"""
A module implementing 'set' methods for bit-fields.
"""

# built-in
from typing import Any, Iterator

# third-party
from vcorelib.io.file_writer import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.struct.methods.bit import (
    bit_field_toggle_method,
    handle_description,
)
from ifgen.struct.methods.fields.common import (
    BitField,
    bit_field_method_slug,
    bit_field_underlying,
    bit_mask_literal,
    possible_array_arg,
)


def bit_field_set_all_method(
    task: GenerateTask,
    writer: IndentedFileWriter,
    header: bool,
    field: dict[str, Any],
    fields: list[BitField],
    alias: str = None,
) -> None:
    """Generate a 'set' method for multiple bit-field."""

    if not header:
        return

    name = field["name"] if not alias else alias

    inner = possible_array_arg(field)
    if inner:
        inner += ", "

    with writer.javadoc():
        writer.write(f"Set all of {name}'s bit fields.")
        handle_description(writer, field)

    # Add field args.
    args = []
    for bit_field in fields:
        args.append(f"{bit_field_underlying(bit_field)} {bit_field['name']}")

    inner += ", ".join(args)
    writer.write(f"inline void set_{name}({inner}){task.method_suffix()}")
    with writer.scope():
        rhs = field["name"] if not alias else alias
        if "index" in inner:
            rhs += "[index]"

        writer.write(f"{field['type']} curr = {rhs};")

        with writer.padding():
            for bit_field in fields:
                for line in bit_field_set_lines(
                    task, bit_field, value=bit_field["name"]
                ):
                    writer.write(line)

        writer.write(f"{rhs} = curr;")


def bit_field_set_lines(
    task: GenerateTask,
    field: BitField,
    lhs: str = "curr",
    value: str = "value",
) -> Iterator[str]:
    """Get lines that perform a bit-field's assignment."""

    mask = bit_mask_literal(field["width"])

    yield f"{lhs} &= ~({mask} << {field['index']}u);"

    val_str = value
    if task.env.is_enum(bit_field_underlying(field)):
        val_str = f"std::to_underlying({val_str})"

    yield f"{lhs} |= ({val_str} & {mask}) << {field['index']}u;"


def bit_field_set_method(
    task: GenerateTask,
    parent: dict[str, Any],
    field: BitField,
    writer: IndentedFileWriter,
    header: bool,
    alias: str = None,
) -> None:
    """Generate a 'set' method for a bit-field."""

    method_slug = bit_field_method_slug(parent, field["name"], alias=alias)
    kind = bit_field_underlying(field)

    # Generate a toggle method for bit fields.
    if field["width"] == 1:
        bit_field_toggle_method(
            task, parent["name"], field, writer, header, method_slug
        )
    else:
        if not header:
            return

        inner = possible_array_arg(parent)
        if inner:
            inner += ", "
        inner += f"{kind} value"

        method = task.cpp_namespace(
            f"set_{method_slug}({inner}){task.method_suffix()}", header=header
        )
        writer.empty()

        if header:
            with writer.javadoc():
                writer.write(f"Set {parent['name']}'s {field['name']} field.")
                handle_description(writer, field)

        writer.write("inline void " + method)
        with writer.scope():
            rhs = parent["name"] if not alias else alias
            if "index" in inner:
                rhs += "[index]"

            writer.write(f"{parent['type']} curr = {rhs};")

            with writer.padding():
                for line in bit_field_set_lines(task, field):
                    writer.write(line)

            writer.write(f"{rhs} = curr;")



"""
A module implementing interfaces for struct-file generation.
"""

# built-in
from contextlib import ExitStack
from typing import Any, Dict, Iterable, Union

# third-party
from vcorelib.io.file_writer import (
    CommentStyle,
    IndentedFileWriter,
    LineWithComment,
)

# internal
from ifgen import PKG_NAME
from ifgen.generation.interface import GenerateTask
from ifgen.struct.methods import struct_methods
from ifgen.struct.methods.fields import bit_fields
from ifgen.struct.source import create_struct_source
from ifgen.struct.stream import struct_stream_methods
from ifgen.struct.test import create_struct_test

__all__ = ["create_struct", "create_struct_test", "create_struct_source"]
FieldConfig = Dict[str, Union[int, str]]


def struct_line(
    name: str,
    value: FieldConfig,
    volatile: bool,
    const: bool,
    array_length: int = None,
) -> LineWithComment:
    """Build a string for a struct-field line."""

    line = f"{value['type']} {name}"
    if array_length is not None:
        line += f"[{name}_length]"

    prefix = "volatile " if volatile else ""
    prefix += "const " if const else ""

    if const:
        line += " = {}"

    return prefix + f"{line};", value.get("description")  # type: ignore


def header_for_type(name: str, task: GenerateTask) -> str:
    """Determine the header file to import for a given type."""

    candidate = task.custom_include(name)
    if candidate:
        return f'"{candidate}"'

    return ""


def struct_includes(task: GenerateTask) -> Iterable[str]:
    """Determine headers that need to be included for a given struct."""

    result = set()
    for config in task.instance["fields"]:
        if "type" in config:
            result.add(header_for_type(config["type"], task))

        # Add includes for bit-fields.
        for bit_field in config.get("fields", []):
            if "type" in bit_field:
                result.add(header_for_type(bit_field["type"], task))

        # Add includes for alternates.
        for alternate in config.get("alternates", []):
            for alternate_bit_field in alternate.get("fields", []):
                if "type" in alternate_bit_field:
                    result.add(
                        header_for_type(alternate_bit_field["type"], task)
                    )

    result.add(f'"../{PKG_NAME}/common.h"')

    return result


def struct_fields(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Generate struct fields."""

    writer.c_comment("Fields.")

    with writer.trailing_comment_lines(style=CommentStyle.C_DOXYGEN) as lines:
        # Fields.
        for field in task.instance["fields"]:
            enforce_expected_size(
                task.env.size(field["type"]) * field.get("array_length", 1),
                field,
                f"{task.name}.{field['name']}",
            )

            if "array_length" in field:
                lines.append(
                    (
                        (
                            f"static constexpr std::size_t "
                            f"{field['name']}_length = "
                            f"{field['array_length']};"
                        ),
                        None,
                    )
                )

            with ExitStack() as stack:
                is_union = field.get("alternates")
                all_fields = [field]
                if is_union:
                    writer.write("union")
                    stack.enter_context(writer.scope(suffix=";"))
                    all_fields.extend(field["alternates"])

                for possible_union in all_fields:
                    if "type" not in possible_union:
                        possible_union["type"] = field["type"]

                    line, comment = struct_line(
                        possible_union["name"],
                        possible_union,
                        possible_union["volatile"],
                        possible_union["const"],
                        array_length=possible_union.get("array_length"),
                    )
                    if is_union:
                        writer.write(
                            line + ("" if not comment else f" /* {comment} */")
                        )
                    else:
                        lines.append((line, comment))

        lines.append(("", None))


def struct_instance(
    task: GenerateTask, writer: IndentedFileWriter, instance: dict[str, Any]
) -> None:
    """Generate struct instances."""

    writer.empty()

    if instance.get("description"):
        with writer.javadoc():
            writer.write(instance["description"])

    type_base = f"{task.name} *"
    writer.write(
        (
            "static "
            + ("volatile " if instance["volatile"] else "")
            + f"{type_base}const "
            f"{instance['name']} = "
            f"reinterpret_cast<{type_base}>({instance['address']});"
        )
    )


def enforce_expected_size(
    size: int, data: dict[str, Any], assert_msg: str
) -> None:
    """Enforce an expected-size field."""

    # If expected size is set, verify it.
    if "expected_size" in data:
        assert data["expected_size"] == size, (
            assert_msg,
            data["expected_size"],
            size,
        )


def create_struct(task: GenerateTask) -> None:
    """Create a header file based on a struct definition."""

    with task.boilerplate(
        includes=struct_includes(task), json=task.instance.get("json", False)
    ) as writer:
        attributes = ["gnu::packed"]
        writer.write(f"struct [[{', '.join(attributes)}]] {task.name}")
        with writer.scope(suffix=";"):
            writer.c_comment("Constant attributes.")
            with writer.trailing_comment_lines(
                style=CommentStyle.C_DOXYGEN
            ) as lines:
                if task.instance["identifier"]:
                    underlying = task.env.config.data["struct"][
                        "id_underlying"
                    ]
                    lines.append(
                        (
                            "static constexpr "
                            f"{underlying} "
                            f"id = {task.protocol().id};",
                            f"{task.name}'s identifier.",
                        )
                    )

                size = task.env.size(task.name)
                enforce_expected_size(size, task.instance, task.name)

                lines.append(
                    (
                        f"static constexpr std::size_t size = {size};",
                        f"{task.name}'s size in bytes.",
                    )
                )

            writer.empty()
            struct_fields(task, writer)

            # Methods.
            writer.c_comment("Methods.")

            if task.instance["methods"]:
                struct_methods(task, writer, True)

            bit_fields(task, writer, True)

        # Add size assertion.
        writer.empty()
        writer.write(
            f"static_assert(sizeof({task.name}) == {task.name}::size);"
        )

        if task.instance["stream"]:
            writer.empty()
            struct_stream_methods(task, writer, True)

        for instance in task.instance.get("instances", []):
            struct_instance(task, writer, instance)



"""
A module implementing a 'device' SVD-element processor.
"""

# built-in
from xml.etree import ElementTree

# third-party
from vcorelib.logging import LoggerType

# internal
from ifgen.svd.model.device import Device
from ifgen.svd.task import SvdProcessingTask


def process_device(
    elem: ElementTree.Element, task: SvdProcessingTask, logger: LoggerType
) -> None:
    """Process a SVD device element."""

    # Assign the device.
    dev = Device.create(elem)
    dev.log(elem, logger)
    task.model.assign_device(dev)

    # Process the CPU metadata.
    cpu = elem.find("cpu")
    if cpu is not None:
        task.process(cpu)

    # Process peripherals.
    peripherals = elem.find("peripherals")
    if peripherals is not None:
        task.process(peripherals)



"""
A module for struct-decoding methods.
"""

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.struct.methods.common import wrapper_method
from ifgen.struct.methods.swap import decode_swapped_method


def decode_method(
    task: GenerateTask, writer: IndentedFileWriter, header: bool
) -> None:
    """Write boilerplate decode method."""

    if header:
        with writer.javadoc():
            writer.write("Update this instance from a buffer.")
            writer.empty()
            writer.write(
                task.command("param[in]", "buffer     Buffer to read.")
            )
            writer.write(
                task.command(
                    "param[in]",
                    "endianness Byte order from decoding elements.",
                )
            )
            writer.write(
                task.command(
                    "return", "              The number of bytes decoded."
                )
            )

    wrapper_method(task, writer, header, is_encode=False)


def struct_decode(
    task: GenerateTask, writer: IndentedFileWriter, header: bool
) -> None:
    """Add a method for decoding structs."""

    with writer.padding():
        decode_swapped_method(task, writer, header)

    decode_method(task, writer, header)



"""
A module implementing 'get' methods for bit-fields.
"""

# built-in
from typing import Any

# third-party
from vcorelib.io.file_writer import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.struct.methods.bit import handle_description
from ifgen.struct.methods.fields.common import (
    BitField,
    bit_field_method_slug,
    bit_field_underlying,
    bit_mask_literal,
    possible_array_arg,
)


def bit_field_get_all_method(
    task: GenerateTask,
    writer: IndentedFileWriter,
    header: bool,
    field: dict[str, Any],
    fields: list[BitField],
    alias: str = None,
) -> None:
    """Generate a 'get' method for multiple bit-field."""

    if not header:
        return

    name = field["name"] if not alias else alias

    inner = possible_array_arg(field)
    if inner:
        inner += ", "

    with writer.javadoc():
        writer.write(f"Get all of {name}'s bit fields.")
        handle_description(writer, field)

    # Add field args.
    args = []
    for bit_field in fields:
        args.append(f"{bit_field_underlying(bit_field)} &{bit_field['name']}")

    inner += ", ".join(args)

    writer.write(f"inline void get_{name}({inner}){task.method_suffix()}")
    with writer.scope():
        rhs = field["name"] if not alias else alias
        if "index" in inner:
            rhs += "[index]"

        writer.write(f"{field['type']} curr = {rhs};")
        writer.empty()

        for bit_field in fields:
            stmt = get_bit_field_statement(task, bit_field, "curr")
            writer.write(f"{bit_field['name']} = {stmt};")


def get_bit_field_statement(
    task: GenerateTask, field: BitField, lhs: str
) -> str:
    """
    Get the arithmetic statement associated with a bit-field get operation.
    """

    kind = bit_field_underlying(field)

    is_flag = field["width"] == 1

    if is_flag:
        stmt = f"{lhs} & (1u << {field['index']}u)"
    else:
        stmt = (
            f"({lhs} >> {field['index']}u) & "
            f"{bit_mask_literal(field['width'])}"
        )

    if task.env.is_enum(kind):
        stmt = f"{kind}({stmt})"

    return stmt


def bit_field_get_method(
    task: GenerateTask,
    parent: dict[str, Any],
    field: BitField,
    writer: IndentedFileWriter,
    header: bool,
    alias: str = None,
) -> None:
    """Generate a 'get' method for a bit-field."""

    if not header:
        return

    inner = possible_array_arg(parent)

    method_slug = bit_field_method_slug(parent, field["name"], alias=alias)
    method = task.cpp_namespace(
        f"get_{method_slug}({inner}){task.method_suffix()}", header=header
    )
    writer.empty()

    with writer.javadoc():
        writer.write(
            (
                f"Get {parent['name']}'s {field['name']} "
                f"{'field' if field['width'] > 1 else 'bit'}."
            )
        )
        handle_description(writer, field)

    line = f"inline {bit_field_underlying(field)} " + method

    lhs = parent["name"] if not alias else alias
    if inner:
        lhs += "[index]"

    writer.write(line)
    with writer.scope():
        writer.write(f"return {get_bit_field_statement(task, field, lhs)};")



"""
A module implementing a 'cpu' SVD-element processor.
"""

# built-in
from xml.etree import ElementTree

# third-party
from vcorelib.logging import LoggerType

# internal
from ifgen.svd.model.cpu import Cpu
from ifgen.svd.task import SvdProcessingTask


def process_cpu(
    elem: ElementTree.Element, task: SvdProcessingTask, logger: LoggerType
) -> None:
    """Process a SVD cpu element."""

    cpu = Cpu()
    cpu.log(elem, logger)
    task.model.assign_cpu(cpu)



# =====================================
# generator=datazen
# version=3.1.4
# hash=089f57617fd119bfbae72f24dfa671c7
# =====================================

"""
This package's command-line entry-point (boilerplate).
"""

# built-in
import argparse
from logging import getLogger
import os
from pathlib import Path
import sys
from typing import List

# third-party
from vcorelib.logging import init_logging, log_time, logging_args

# internal
from ifgen import DESCRIPTION, VERSION
from ifgen.app import add_app_args, entry


def main(argv: List[str] = None) -> int:
    """Program entry-point."""

    result = 0

    # fall back on command-line arguments
    command_args = sys.argv
    if argv is not None:
        command_args = argv

    # initialize argument parsing
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {VERSION}",
    )
    logging_args(parser)
    parser.add_argument(
        "-C",
        "--dir",
        default=Path.cwd(),
        dest="dir",
        type=Path,
        help="execute from a specific directory",
    )
    starting_dir = Path.cwd()

    add_app_args(parser)

    # parse arguments and execute the requested command
    try:
        args = parser.parse_args(command_args[1:])
        args.version = VERSION
        args.dir = args.dir.resolve()

        # initialize logging
        init_logging(
            args, default_format="%(name)-36s - %(levelname)-6s - %(message)s"
        )

        # change to the specified directory
        os.chdir(args.dir)

        # run the application
        with log_time(getLogger(__name__), "Command"):
            result = entry(args)
    except SystemExit as exc:
        result = 1
        if exc.code is not None and isinstance(exc.code, int):
            result = exc.code

    # return to starting dir
    os.chdir(starting_dir)

    return result



"""
A module implementing interfaces for byte-swapping method generation.
"""

# built-in
from contextlib import ExitStack
from typing import Any

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask


def no_swap(
    field: dict[str, Any],
    is_decode: bool,
    task: GenerateTask,
    writer: IndentedFileWriter,
    is_array: bool,
) -> None:
    """Encode or decode a single-byte type."""

    name = field["name"]
    if is_array:
        name += "[i]"

    kind = field["type"]

    is_enum = task.env.is_enum(kind)

    if is_decode:
        line = f"{name} = "
        arg = "buf[idx++]"
        if is_enum:
            arg = f"{kind}({arg})"
        else:
            arg = f"std::to_integer<{kind}>({arg})"

        writer.write(line + arg + ";")
    else:
        writer.write(f"buf[idx++] = std::byte({name});")


def swap_struct(
    field: dict[str, Any],
    is_decode: bool,
    task: GenerateTask,
    writer: IndentedFileWriter,
    is_array: bool,
) -> None:
    """Perform a byte swap for a struct type."""

    del task

    name = field["name"]
    if is_array:
        name += "[i]"

    kind = field["type"]

    writer.write(
        f"idx += {name}.{'decode' if is_decode else 'encode'}_swapped("
    )
    pointer = f"{kind}::Buffer *"
    if is_decode:
        pointer = "const " + pointer

    arg = f"reinterpret_cast<{pointer}>"
    arg += "(&(*buffer)[idx])"
    writer.write(f"    {arg});")


def assignment(writer: IndentedFileWriter, lhs: str, rhs: str) -> None:
    """Write an assignment line."""

    line = f"{lhs} = {rhs};"

    if len(line) + 4 > 79:
        start = lhs + " ="
        writer.write(start)
        writer.write("    " + rhs + ";")
    else:
        writer.write(line)


def to_integral(kind: str) -> str:
    """Convert certain types to their corresponding integral types."""

    lazy = {"float": "uint32_t", "double": "uint64_t"}
    return lazy.get(kind, kind)


def swap_enum(
    field: dict[str, Any],
    is_decode: bool,
    task: GenerateTask,
    writer: IndentedFileWriter,
    is_array: bool,
) -> None:
    """Perform a byte swap for an enumeration type."""

    underlying = task.env.get_enum(field["type"]).primitive + "_t"

    name = field["name"]
    if is_array:
        name += "[i]"

    if is_decode:
        lhs = name
        underlying = to_integral(underlying)
        rhs = (
            f"{field['type']}(std::byteswap(*reinterpret_cast<const "
            f"{underlying} *>(&buf[idx])))"
        )

    else:
        lhs = f"*reinterpret_cast<{underlying} *>(&buf[idx])"
        underlying = to_integral(underlying)
        rhs = f"std::byteswap(std::to_underlying({name}))"

    assignment(writer, lhs, rhs)


def encode_primitive_swap(
    field: dict[str, Any], writer: IndentedFileWriter, is_array: bool
) -> None:
    """Encode a primitive-sized element by swapping byte order."""

    underlying = field["type"]
    integral = to_integral(underlying)

    name = field["name"]
    if is_array:
        name += "[i]"

    rhs = "std::byteswap("
    if field["type"] == integral:
        lhs = f"*reinterpret_cast<{underlying} *>(&buf[idx])"
        rhs += f"{name})"
    else:
        lhs = f"*reinterpret_cast<{integral} *>(&buf[idx])"
        rhs += f"reinterpret_cast<const {integral} &>({name}))"

    assignment(writer, lhs, rhs)


def decode_primitive_swap(
    field: dict[str, Any], writer: IndentedFileWriter, is_array: bool
) -> None:
    """Decode a primitive-sized element by swapping byte order."""

    lhs = f"{field['name']}"
    if is_array:
        lhs += "[i]"

    underlying = field["type"]
    integral = to_integral(underlying)
    reinterp = f"reinterpret_cast<const {integral} *>"

    if integral == underlying:
        rhs = f"std::byteswap(*{reinterp}" f"(&buf[idx]))"
        assignment(writer, lhs, rhs)
    else:
        assignment(
            writer, "auto val", f"std::byteswap(*{reinterp}(&buf[idx]))"
        )
        assignment(
            writer, lhs, f"*reinterpret_cast<const {underlying} *>(&val)"
        )


def swap_fields(
    task: GenerateTask, writer: IndentedFileWriter, is_decode: bool = True
) -> None:
    """Perform byte swaps on individual struct fields."""

    writer.write("std::size_t idx = 0;")
    writer.write("auto buf = buffer->data();")

    for field in task.instance["fields"]:
        writer.empty()

        name = field["name"]
        kind = field["type"]
        writer.c_comment(f"{kind} {name}")

        with ExitStack() as stack:
            is_array = "array_length" in field
            if is_array:
                array_cmp = task.cpp_namespace(f"{name}_length")
                writer.write(f"for (std::size_t i = 0; i < {array_cmp}; i++)")
                stack.enter_context(writer.scope())

            size = task.env.size(kind)

            # Handle padding.
            if field["padding"]:
                writer.c_comment(f"Advance for padding field '{name}'.")
                writer.write(f"idx += {size};")
                continue

            if size == 1:
                no_swap(field, is_decode, task, writer, is_array)
            elif task.env.is_struct(kind):
                swap_struct(field, is_decode, task, writer, is_array)
            elif task.env.is_enum(kind):
                swap_enum(field, is_decode, task, writer, is_array)
                writer.write(f"idx += {size};")
            else:
                if is_decode:
                    decode_primitive_swap(field, writer, is_array)
                else:
                    encode_primitive_swap(field, writer, is_array)
                writer.write(f"idx += {size};")

    writer.empty()
    writer.write("return idx;")


def decode_swapped_method(
    task: GenerateTask, writer: IndentedFileWriter, header: bool
) -> None:
    """Generate a struct-decode method that uses swapped byte order."""

    if header:
        with writer.javadoc():
            writer.write("Decode using byte-order swapped from native.")
            writer.empty()
            writer.write(task.command("param[in]", "buffer Buffer to read."))
            writer.write(
                task.command(
                    "return", "          The number of bytes decoded."
                )
            )

    writer.write(
        f"std::size_t {task.cpp_namespace('decode_swapped', header=header)}"
        "(const Buffer *buffer)" + (";" if header else "")
    )

    if header:
        return

    with writer.scope():
        swap_fields(task, writer)


def encode_swapped_method(
    task: GenerateTask, writer: IndentedFileWriter, header: bool
) -> None:
    """Generate a struct-encode method that uses swapped byte order."""

    if header:
        with writer.javadoc():
            writer.write("Encode using byte-order swapped from native.")
            writer.empty()
            writer.write(task.command("param[out]", "buffer Buffer to write."))
            writer.write(
                task.command(
                    "return", "           The number of bytes encoded."
                )
            )

    method = task.cpp_namespace("encode_swapped", header=header)
    writer.write(
        f"std::size_t {method}(Buffer *buffer) const" + (";" if header else "")
    )

    if header:
        return

    with writer.scope():
        swap_fields(task, writer, is_decode=False)



"""
A module implementing a data model for ARM CMSIS-SVD 'enumeratedValue' data.
"""

# built-in
from dataclasses import dataclass
from typing import Iterable, Optional
from xml.etree import ElementTree

# internal
from ifgen.svd.model.derived import DerivedMixin
from ifgen.svd.string import StringKeyVal, StringKeyValueMixin


@dataclass
class EnumeratedValue(StringKeyValueMixin):
    """A container for enumerated-value information."""

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        return [
            StringKeyVal("name", False),
            StringKeyVal("description", False),
            # Some kind of parsing needed for this field maybe.
            StringKeyVal("value", False),
            StringKeyVal("isDefault", False),
        ]


EnumValueMap = dict[str, EnumeratedValue]


@dataclass
class EnumeratedValues(DerivedMixin):
    """A container for enum information."""

    derived_from: Optional["EnumeratedValues"]
    enum: EnumValueMap

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        return [
            StringKeyVal("name", False),
            StringKeyVal("headerEnumName", False),
            StringKeyVal("usage", False),
        ]


def get_enum(enumerated_values: ElementTree.Element) -> EnumeratedValues:
    """Get register clusters."""

    # Handle values.
    vals = {}
    for enum in enumerated_values.iterfind("enumeratedValue"):
        val_inst = EnumeratedValue.create(enum)
        vals[val_inst.raw_data["name"]] = val_inst

    # No handling for derived enumerations yet.
    derived = enumerated_values.attrib.get("derivedFrom")
    assert derived is None, derived

    return EnumeratedValues.create(enumerated_values, None, vals)



"""
A module implementing an SVD-processing task interface.
"""

# built-in
from dataclasses import dataclass
from logging import getLogger
from pathlib import Path
from typing import Callable, Iterable, Iterator
from xml.etree import ElementTree

# third-party
from vcorelib.io import ARBITER
from vcorelib.logging import LoggerType
from vcorelib.paths import rel

# internal
from ifgen.config.svd import SvdConfig
from ifgen.svd.group import handle_group, peripheral_groups
from ifgen.svd.model import SvdModel

TagProcessor = Callable[
    [ElementTree.Element, "SvdProcessingTask", LoggerType], None
]
TagProcessorMap = dict[str, TagProcessor]
TAG_PROCESSORS: TagProcessorMap = {}


def filter_includes(
    config: SvdConfig,
    model: SvdModel,
    paths: Iterable[Path],
    filtered: dict[str, str],
) -> Iterator[str]:
    """Filter includes based on configuration."""

    config_data = config.data.setdefault("devices", {}).setdefault(
        model.device_name.lower(), {"ignore_peripherals": []}
    )

    for path in paths:
        item = str(path)
        reason = None

        # Determine a possible reason to filter this include.
        for ignore in config_data["ignore_peripherals"]:
            if path.parts[0] == ignore["name"]:
                reason = ignore["reason"]

        if reason is not None:
            filtered[item] = reason
        else:
            yield item


@dataclass
class SvdProcessingTask:
    """A container for SVD-processing state."""

    model: SvdModel
    min_enum_width: int

    def process(self, elem: ElementTree.Element) -> None:
        """Process a single element."""
        TAG_PROCESSORS[elem.tag](elem, self, getLogger(elem.tag))

    @staticmethod
    def svd(path: Path, min_enum_width: int) -> "SvdProcessingTask":
        """Process a single SVD file."""

        task = SvdProcessingTask(SvdModel({}), min_enum_width)
        task.process(ElementTree.parse(path).getroot())
        return task

    def generate_configs(self, path: Path, config: SvdConfig) -> None:
        """Generate output configuration files."""

        path.mkdir(exist_ok=True, parents=True)

        includes: set[Path] = set()

        # Organize peripherals into groups based on ones derived from others
        # and process them.
        for group in peripheral_groups(self.model.peripherals).values():
            output_dir = path.joinpath(group.root.base_name())
            output_dir.mkdir(exist_ok=True)
            handle_group(output_dir, group, includes, self.min_enum_width)

        # Write metadata that doesn't currently get used for generation.
        meta = self.model.metadata()

        # Indicate includes that were filtered out in metadata.
        filtered: dict[str, str] = {}
        meta["filtered_includes"] = filtered

        ARBITER.encode(
            path.joinpath("ifgen.yaml"),
            {
                "includes": sorted(  # type: ignore
                    filter_includes(
                        config,
                        self.model,
                        (rel(x.resolve(), base=path) for x in includes),
                        filtered,
                    )
                ),
                "namespace": self.model.namespace(),  # type: ignore
                "struct": {
                    "stream": False,
                    "codec": False,
                    "methods": False,
                    "unit_test": False,
                    "identifier": False,
                },
                "enum": {"use_map": False, "identifier": False},
            },
        )
        ARBITER.encode(path.joinpath("metadata.json"), meta)



"""
A module implementing a generation-environment interface.
"""

# built-in
from enum import StrEnum
from pathlib import Path
from shutil import rmtree
from typing import Any

# third-party
from runtimepy.codec.protocol import Protocol
from runtimepy.codec.system import TypeSystem
from runtimepy.enum import RuntimeEnum
from vcorelib.logging import LoggerMixin
from vcorelib.paths import normalize, rel

# internal
from ifgen import PKG_NAME
from ifgen.config import Config
from ifgen.environment.field import process_field
from ifgen.environment.padding import PaddingManager, type_string
from ifgen.paths import combine_if_not_absolute


class Generator(StrEnum):
    """An enumeration declaring all valid kinds of generators."""

    STRUCTS = "structs"
    ENUMS = "enums"
    IFGEN = PKG_NAME


def runtime_enum_data(data: dict[str, Any]) -> dict[str, int]:
    """Get runtime enumeration data."""

    result = {}

    curr_value = 0

    for key, value in data.items():
        if value is None or "value" not in value:
            result[key] = curr_value
            curr_value += 1
        else:
            result[key] = value["value"]
            if value["value"] >= curr_value:
                curr_value = value["value"] + 1

    return result


class IfgenEnvironment(LoggerMixin):
    """A class for managing stateful information while generating outputs."""

    def __init__(self, root: Path, config: Config) -> None:
        """Initialize this instance."""

        super().__init__()
        self.root_path = root
        self.config = config

        self.source = combine_if_not_absolute(
            self.root_path, normalize(*self.config.data["source_dir"])
        )
        self.output = combine_if_not_absolute(
            self.source, normalize(*self.config.data["output_dir"])
        )
        self.test_dir = combine_if_not_absolute(
            self.source, normalize(*self.config.data["test_dir"])
        )

        self.generated: set[Path] = set()

        # Create output directories.
        for subdir in Generator:
            for path in [self.output, self.test_dir]:
                dest = path.joinpath(subdir)
                rmtree(dest, ignore_errors=True)
                dest.mkdir(parents=True, exist_ok=True)

        self.types = TypeSystem(*self.config.data["namespace"])
        self.padding = PaddingManager()
        self._register_enums()
        self._register_structs()

    def _register_enums(self) -> None:
        """Register configuration enums."""

        for name, enum in self.config.data.get("enums", {}).items():
            self.types.enum(
                name,
                runtime_enum_data(enum["enum"]),
                *enum["namespace"],
                primitive=type_string(enum["underlying"]),
            )

            self.logger.info(
                "Registered enum '%s'.",
                self.types.root_namespace.delim.join(
                    enum["namespace"] + [name]
                ),
            )

    def _register_structs(self) -> None:
        """Register configuration structs."""

        for name, struct in self.config.data.get("structs", {}).items():
            namespace = [*struct["namespace"]]
            self.types.register(name, *namespace)

            field_groups = []

            self.padding.reset()
            for field in struct["fields"]:
                padding = list(
                    process_field(
                        name, self.padding, self.types, field, namespace
                    )
                )
                if padding:
                    field_groups.append(padding)
                field_groups.append([field])

            # Re-assign fields structure.
            struct["fields"] = []
            for group in field_groups:
                for field in group:
                    struct["fields"].append(field)

            self.logger.info(
                "Registered struct '%s' (%d bytes).",
                self.types.root_namespace.delim.join(
                    struct["namespace"] + [name]
                ),
                self.types.size(name, *struct["namespace"]),
            )

    def make_path(
        self,
        name: str,
        generator: Generator,
        from_output: bool = False,
        track: bool = True,
    ) -> Path:
        """Make part of a task's path."""

        result = Path(str(generator), f"{name}.h")

        if from_output:
            result = self.output.joinpath(result)

        if track:
            self.generated.add(result)

        return result

    def rel_include(self, name: str, generator: Generator) -> Path:
        """Get an include path to a generated output."""

        return rel(self.output, base=self.source).joinpath(
            self.make_path(name, generator, track=False)
        )

    def make_test_path(self, name: str, generator: Generator) -> Path:
        """Make a path to an interface's unit-test suite."""

        result = self.test_dir.joinpath(str(generator), f"test_{name}.cc")
        self.generated.add(result)
        return result

    def get_protocol(self, name: str, exact: bool = False) -> Protocol:
        """Get the protocol instance for a given struct."""

        return self.types.get_protocol(
            name, self.config.data["structs"].get("namespace", []), exact=exact
        )

    def is_struct(self, name: str) -> bool:
        """Determine if a field is a struct or not."""

        try:
            self.get_protocol(type_string(name))
            return True
        except KeyError:
            return False

    def size(self, type_name: str, exact: bool = False) -> int:
        """Get the size of a given type."""
        return self.types.size(type_string(type_name), exact=exact)

    def get_enum(self, name: str, exact: bool = False) -> RuntimeEnum:
        """Get a runtime enum instance for a given enumeration."""

        return self.types.get_enum(
            name, *self.config.data["enums"].get("namespace", []), exact=exact
        )

    def is_enum(self, name: str, exact: bool = False) -> bool:
        """Determine if a field is an enumeration or not."""

        try:
            self.get_enum(type_string(name), exact=exact)
            return True
        except KeyError:
            return False



"""
A module for generating enumeration header files.
"""

# built-in
from typing import Dict, Optional, Union

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.enum.common import enum_to_string_function, string_to_enum_function
from ifgen.generation.interface import GenerateTask
from ifgen.generation.json import to_json_method

EnumConfig = Optional[Dict[str, Union[int, str]]]


def enum_line(name: str, value: EnumConfig) -> str:
    """Build a string representing a line in an enumeration."""

    line = name

    if value and value.get("value"):
        line += f" = {value['value']}"

    if value and value.get("description"):
        line += f" /*!< {value['description']} */"

    return line


def enum_header(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Create a header file for an enumeration."""

    writer.write(f"enum class {task.name} : {task.instance['underlying']}")
    with writer.scope(suffix=";"):
        writer.join(
            *(
                enum_line(enum, value)
                for enum, value in task.instance.get("enum", {}).items()
            )
        )

    writer.write(
        (
            f"static_assert(sizeof({task.name}) == "
            f"{task.env.size(task.name)});"
        )
    )

    runtime = task.enum()

    writer.empty()

    if task.instance["identifier"]:
        writer.write(
            (
                "static constexpr "
                f"{task.env.config.data['enum']['id_underlying']} "
                f"{task.name}_id = {runtime.id};"
            )
        )
        writer.empty()

    enum_to_string_function(
        task, writer, task.instance["use_map"], definition=True
    )

    writer.empty()
    string_to_enum_function(
        task, writer, task.instance["use_map"], definition=True
    )

    if task.instance["json"]:
        to_json_method(
            task,
            writer,
            runtime.asdict(),
            dumps_indent=task.instance["json_indent"],
            definition=True,
            inline=True,
        )



"""
A module implementing a unit-test output generator for structs.
"""

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.generation.test import (
    unit_test_boilerplate,
    unit_test_main,
    unit_test_method,
    unit_test_method_name,
)


def assert_line(writer: IndentedFileWriter, data: str) -> None:
    """Write an assert line to the file."""
    writer.write(f"assert({data});")


def unit_test_stream_tests(
    task: GenerateTask, writer: IndentedFileWriter
) -> None:
    """Generate unit tests for stream interfaces."""

    writer.c_comment("Test stream interactions.")

    writer.write("static constexpr std::size_t len = 10;")
    writer.write(f"byte_array<{task.name}::size * len> streambuf;")
    writer.write(f"using TestSpan = byte_span<{task.name}::size * len>;")
    writer.write("auto stream = byte_spanstream(TestSpan(streambuf));")

    with writer.padding():
        writer.write("stream << dst;")
        writer.write("stream.seekg(0);")

    writer.write(f"{task.name} from_stream;")
    writer.write("stream >> from_stream;")


def unit_test_basic_method(
    task: GenerateTask, writer: IndentedFileWriter
) -> None:
    """
    Implement a simple encode-and-decode scenario with endianness as a
    function argument.
    """

    with unit_test_method("encode_decode_basic", task, writer):
        nspaced = task.name
        writer.write(f"{nspaced} src = {'{}'};")

        if task.instance["methods"]:
            assert_line(writer, f"src.span().size() == {nspaced}::size")
        else:
            writer.write("(void)src;")

        if task.instance["codec"]:
            with writer.padding():
                writer.write("src.swap();")

            writer.c_comment("Eventually, we could assign member values here.")

            with writer.padding():
                writer.write(f"{nspaced}::Buffer buffer;")
                assert_line(
                    writer,
                    f"src.encode(&buffer, endianness) == {nspaced}::size",
                )

            writer.write(f"{nspaced} dst;")
            assert_line(
                writer, f"dst.decode(&buffer, endianness) == {nspaced}::size"
            )
            assert_line(writer, "src == dst")

            writer.empty()
            writer.c_comment("Verify the values transferred.")
        else:
            writer.write("(void)endianness;")
            if task.instance["stream"]:
                writer.empty()
                writer.write(f"{nspaced} dst = {'{}'};")

        if task.instance["stream"]:
            writer.empty()
            unit_test_stream_tests(task, writer)


def unit_test_body(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Implement a unit test for a struct."""

    for method in ["encode_decode_basic"]:
        for arg in [
            "std::endian::native",
            "std::endian::little",
            "std::endian::big",
        ]:
            name = unit_test_method_name(method, task)
            writer.write(f"{name}({arg});")

        writer.empty()

    if task.instance["json"]:
        writer.c_comment("Attempt to decode this?")
        writer.write(f"std::cout << {task.name}::json();")


def create_struct_test(task: GenerateTask) -> None:
    """Create a unit test for the enum string-conversion methods."""

    if task.instance["unit_test"]:
        with unit_test_boilerplate(task, main=False) as writer:
            unit_test_basic_method(task, writer)

            writer.empty()

            with unit_test_main(task, writer):
                unit_test_body(task, writer)



"""
A module implementing a data model for ARM CMSIS-SVD 'device' data.
"""

# built-in
from dataclasses import dataclass
from typing import Any, Optional
from xml.etree import ElementTree

# internal
from ifgen.svd.model.cpu import Cpu
from ifgen.svd.model.device import Device
from ifgen.svd.model.peripheral import Peripheral


@dataclass
class SvdModel:
    """A model for SVD data."""

    peripherals: dict[str, Peripheral]
    device: Optional[Device] = None
    cpu: Optional[Cpu] = None

    @property
    def device_name(self) -> str:
        """Get the SVD device name."""
        assert self.device is not None
        return self.device.raw_data["name"]

    def namespace(self) -> list[str]:
        """Get a namespace for this SVD device."""
        return [x.upper() for x in self.device_name.split("_")]

    def metadata(self) -> dict[str, Any]:
        """Get device and CPU metadata."""

        result: dict[str, Any] = {}

        if self.device is not None:
            result["device"] = self.device.raw_data
        if self.cpu is not None:
            result["cpu"] = self.cpu.raw_data

        for name, peripheral in self.peripherals.items():
            data = {
                "interrupts": [x.raw_data for x in peripheral.interrupts],
                "address_blocks": [
                    x.raw_data for x in peripheral.address_blocks
                ],
            }

            # Add alternate group metadata.
            groups = peripheral.register_groups()
            if groups:
                data["register_groups"] = {  # type: ignore
                    key: [x.name for x in value]
                    for key, value in groups.items()
                }

            result[name] = data

        return result

    def assign_device(self, device: Device) -> None:
        """Assign a device instance."""
        assert self.device is None, self.device
        self.device = device

    def assign_cpu(self, cpu: Cpu) -> None:
        """Assign a cpu instance."""
        assert self.cpu is None, self.cpu
        self.cpu = cpu

    def register_peripheral(
        self, elem: ElementTree.Element, peripheral: Peripheral
    ) -> None:
        """Register a peripheral."""

        name = peripheral.raw(elem)["name"]
        assert name not in self.peripherals, f"Duplicate peripheral '{name}'!"
        self.peripherals[name] = peripheral



# =====================================
# generator=datazen
# version=3.1.4
# hash=347b61ef0ea6ed99b2f085b2d17fb232
# =====================================

"""
A module aggregating package commands.
"""

# built-in
from typing import List as _List
from typing import Tuple as _Tuple

# third-party
from vcorelib.args import CommandRegister as _CommandRegister

# internal
from ifgen.commands.gen import add_gen_cmd
from ifgen.commands.svd import add_svd_cmd


def commands() -> _List[_Tuple[str, str, _CommandRegister]]:
    """Get this package's commands."""

    return [
        (
            "gen",
            "generate interfaces",
            add_gen_cmd,
        ),
        (
            "svd",
            "process CMSIS-SVD files",
            add_svd_cmd,
        ),
        ("noop", "command stub (does nothing)", lambda _: lambda _: 0),
    ]
